@@ -7,6 +7,7 @@ use tracing::trace;
 
 use self::scope::Scope;
 use super::rename_map::RenameMap;
+use crate::pass::compute_char_freq::CharFreqInfo;
 
 mod scope;
 
@@ -17,11 +18,16 @@ pub(super) struct Analyzer {
 }
 
 impl Analyzer {
-    pub(super) fn into_rename_map(mut self, preserved: &AHashSet<Id>) -> AHashMap<Id, JsWord> {
+    pub(super) fn into_rename_map(
+        mut self,
+        freq: &CharFreqInfo,
+        preserved: &AHashSet<Id>,
+    ) -> AHashMap<Id, JsWord> {
         let mut map = RenameMap::default();
 
         let preserved_symbols = preserved.iter().cloned().map(|v| v.0).collect();
-        self.scope.rename(&mut map, preserved, &preserved_symbols);
+        self.scope
+            .rename(freq, &mut map, preserved, &preserved_symbols);
 
         map.into_map()
     }
@@ -112,6 +118,20 @@ impl Visit for Analyzer {
         })
     }
 
+    fn visit_class_method(&mut self, f: &ClassMethod) {
+        f.key.visit_with(self);
+
+        self.with_scope(|v| {
+            f.function.visit_with(v);
+        })
+    }
+
+    fn visit_constructor(&mut self, f: &Constructor) {
+        self.with_scope(|v| {
+            f.visit_children_with(v);
+        })
+    }
+
     fn visit_default_decl(&mut self, d: &DefaultDecl) {
         match d {
             DefaultDecl::Class(c) => {
